/****************************************************************************
**
** Copyright (C) 2013 Digia Plc and/or its subsidiary(-ies).
** Contact: http://www.qt-project.org/legal
**
** This file is part of the documentation of the Qt Toolkit.
**
** $QT_BEGIN_LICENSE:BSD$
** You may use this file under the terms of the BSD license as follows:
**
** "Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are
** met:
**   * Redistributions of source code must retain the above copyright
**     notice, this list of conditions and the following disclaimer.
**   * Redistributions in binary form must reproduce the above copyright
**     notice, this list of conditions and the following disclaimer in
**     the documentation and/or other materials provided with the
**     distribution.
**   * Neither the name of Digia Plc and its Subsidiary(-ies) nor the names
**     of its contributors may be used to endorse or promote products derived
**     from this software without specific prior written permission.
**
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
** "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
** LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
** A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
** OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
** SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
** LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
** DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
** THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
** OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
**
** $QT_END_LICENSE$
**
****************************************************************************/

#include "openglwindow.h"

#include <QtCore/QCoreApplication>
#include <QtGui/QGuiApplication>
#include <QtGui/QScreen>

#include <QtGui/QOpenGLContext>
#include <QtGui/QOpenGLPaintDevice>
#include <QtGui/QPainter>
#include <QStandardPaths>

#include "seamless3d.h"
#include "render-ogl.h"

QString downloadLocation;

unsigned int window_width = 1280, window_height = 720;
const int size = window_width * window_height;

unsigned int subwin_width = 256, subwin_height = 256;
const int subwin_size=subwin_width*subwin_height;
unsigned char* pixels = new unsigned char[subwin_size*3];

Ptr<Mesh> g_mesh;


//! [1]
OpenGLWindow::OpenGLWindow(QWindow *parent)
    : QWindow(parent)
    , m_update_pending(false)
    , m_animating(false)
    , m_context(0)
    , m_device(0)
{
    setSurfaceType(QWindow::OpenGLSurface);
}
//! [1]

OpenGLWindow::~OpenGLWindow()
{
    delete m_device;
}
//! [2]
void OpenGLWindow::render(QPainter *painter)
{
    Q_UNUSED(painter);
}

void OpenGLWindow::renderLater()
{
    if (!m_update_pending) {
        m_update_pending = true;
        QCoreApplication::postEvent(this, new QEvent(QEvent::UpdateRequest));
    }
}

bool OpenGLWindow::event(QEvent *event)
{
    switch (event->type()) {
    case QEvent::UpdateRequest:
        m_update_pending = false;
        renderNow();
        return true;
    case QEvent::Resize:
        window_width = width();
        window_height = height();
        return QWindow::event(event);

    default:
        return QWindow::event(event);
    }
}

void OpenGLWindow::exposeEvent(QExposeEvent *event)
{
    Q_UNUSED(event);

    if (isExposed())
        renderNow();
}
//! [3]

//! [4]
void OpenGLWindow::renderNow()
{
    if (!isExposed())
        return;

    bool needsInitialize = false;

    if (!m_context) {
        m_context = new QOpenGLContext(this);
        m_context->setFormat(requestedFormat());
        m_context->create();

        needsInitialize = true;
    }

    m_context->makeCurrent(this);

    if (needsInitialize) {
        //initializeOpenGLFunctions();
        initialize();
    }

    render();

    m_context->swapBuffers(this);

    if (m_animating)
        renderLater();
}
//! [4]

//! [5]
void OpenGLWindow::setAnimating(bool animating)
{
    m_animating = animating;

    if (animating)
        renderLater();
}
//! [5]
#include <QMouseEvent>

#include <math.h>

QQuaternion rotation;
Vec2i g_mousePos;

bool g_viewRotDragMode = false;
Vec2i g_rightDragBegin;
Rotation sceneRotation;
Rotation g_beginViewRot;

void OpenGLWindow::mousePressEvent(QMouseEvent *e)
{
    g_viewRotDragMode = true;
    g_rightDragBegin.x = e->localPos().x();
    g_rightDragBegin.y = e->localPos().y();
    g_beginViewRot = sceneRotation;
}


void OpenGLWindow::mouseReleaseEvent(QMouseEvent *e)
{
    g_viewRotDragMode = false;
}

void OpenGLWindow::mouseMoveEvent(QMouseEvent *e)
{
    Vec2i pos;
    g_mousePos.x = e->localPos().x();
    g_mousePos.y = e->localPos().y();

    if(e->buttons() & Qt::MiddleButton){
        Vec2f posA2f = Vec2f(g_rightDragBegin.x - ((int)window_width / 2), -g_rightDragBegin.y + ((int)window_height / 2));
        Vec2f posB2f = Vec2f(g_mousePos.x - ((int)window_width / 2), -g_mousePos.y + ((int)window_height / 2));
        float a1 = posA2f.getAngle();
        float a2 = posB2f.getAngle();
        float a3 = a2 - a1;
        sceneRotation = g_beginViewRot * Rotation(0, 0, 1, a3);
    }
    else if(e->buttons() & Qt::RightButton){
        if(g_viewRotDragMode){

            Vec2i diffi = g_mousePos - g_rightDragBegin;
            Vec2f diff = Vec2f(diffi.x, diffi.y);
            float distance = diff.length();

            float a = diff.getAngle2();
            Vec3f axis = Vec3f(0, 1, 0);
            axis = Rotation(0, 0, 1, -a) * axis;
            //axis.z = -axis.z;

            sceneRotation = g_beginViewRot * Rotation(axis, distance*.005);

        }
    }
}

void OpenGLWindow::timerEvent(QTimerEvent *)
{
}
void initLookup();
String readFile(String filename);
void createSmlsFromString(M_Node& nodes, const String& str);

void setUpMesh(){
    initLookup();
    String str = readFile(downloadLocation.toStdString().c_str());
    M_Node nodes;
    createSmlsFromString(nodes, str);

    if(nodes.length == 0){
        return;
    }

    Mesh* mesh = dynamic_cast<Mesh*>(nodes[0].p);
    g_mesh = mesh;
    mesh->render = new Render;
    mesh->render->convertToRenderFormatCNC(mesh->coord, mesh->normal, mesh->color, mesh->coordIndex);
}


void OpenGLWindow::initialize()
{
    initializeOpenGLFunctions();

    glClearColor(1, 1, 1, 1);

    Render::initShader(CNCVertex::program, ":/vshader.glsl", ":/fshader.glsl");

    // Enable depth buffer
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);

    //Enable back face culling
    glEnable(GL_CULL_FACE);

    m_coord = CNCVertex::program.attributeLocation("a_coord");
    m_normal = CNCVertex::program.attributeLocation("a_normal");
    m_color = CNCVertex::program.attributeLocation("a_color");
    m_matrixUniform = CNCVertex::program.uniformLocation("mvp_matrix");


    setUpMesh();
}
//! [4]

float m_frame = 0;
QMatrix4x4 projection;
extern Rotation sceneRotation;

void OpenGLWindow::render()
{

    if (!m_device){
        m_device = new QOpenGLPaintDevice;
    }
    CNCVertex::program.bind();



    QPainter painter(m_device);
    render(&painter);

    float w = width();
    float h = height();

    const qreal retinaScale = devicePixelRatio();
    glViewport(0, 0, w * retinaScale, h * retinaScale);

    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);
    //glDepthFunc(GL_LESS);
    //glDepthRangef(0.0,100.0);

    //glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);


    //glMatrixMode(GL_PROJECTION);
    double aspect = w / h;

    // Set near plane to 3.0, far plane to 7.0, field of view 45 degrees
    //const qreal zNear = .25, zFar = 10.0, fov = 90.0;
    //const qreal zNear = .25, zFar = 100.0, fov = 2.0;

    // Reset projection
    projection.setToIdentity();

    // Set perspective projection
    //projection.perspective(fov, aspect, zNear, zFar);

    projection.ortho(-aspect, aspect, -1, 1, -3, 100);
    //glMatrixMode(GL_MODELVIEW);


    QMatrix4x4 matrix;
    matrix.setToIdentity();


    //rotation = QQuaternion::fromAxisAndAngle(0, .707107, .707107, 10.0f * m_frame / screen()->refreshRate());
    //rotation = QQuaternion::fromAxisAndAngle(sceneRotation.x, sceneRotation.y, sceneRotation.z, sceneRotation.angle);

    //matrix.rotate(rotation);

   //matrix.lookAt(QVector3D(0.0f, 0.0f, 3.0f), QVector3D(0, 0.0f, 0.0f), QVector3D(0, 1, 0));

    Matrix rotMatrix = rotationMatrix(sceneRotation);
    //Matrix sc = scaleMatrix(Vec3f(1, 1 , -1));
    //rotMatrix = rotMatrix * sc;


    QMatrix4x4& qRotMatrix = *reinterpret_cast<QMatrix4x4*>(&rotMatrix);

    matrix = matrix * qRotMatrix;

    //matrix.scale(6, 6, 6);
    //matrix.translate(0.0, -1.65, 0.0);



    // Set modelview-projection matrix
    CNCVertex::program.setUniformValue("mvp_matrix", projection * matrix);

    QMatrix4x4 normalMatrix;
    normalMatrix.rotate(rotation);

    CNCVertex::program.setUniformValue("u_normalMatrix", qRotMatrix);




    // Tell OpenGL which VBOs to use
    g_mesh->render->arrayBuf.bind();
    g_mesh->render->indexBuf.bind();

    // Offset for coord
    quintptr offset = 0;

    // Tell OpenGL programmable pipeline how to locate vertex coord data
    int vertexLocation = CNCVertex::program.attributeLocation("a_coord");
    CNCVertex::program.enableAttributeArray(vertexLocation);
    CNCVertex::program.setAttributeBuffer(vertexLocation, GL_FLOAT, offset, 3, sizeof(CNCVertex));


    // Offset for texture coordinate
    offset += sizeof(QVector3D);

    // Tell OpenGL programmable pipeline how to locate vertex texture coordinate data
    int normalLocation = CNCVertex::program.attributeLocation("a_normal");
    CNCVertex::program.enableAttributeArray(normalLocation);
    CNCVertex::program.setAttributeBuffer(normalLocation, GL_FLOAT, offset, 3, sizeof(CNCVertex));


    // Offset for texture coordinate
    offset += sizeof(QVector3D);

    // Tell OpenGL programmable pipeline how to locate vertex texture coordinate data
    int colorLocation = CNCVertex::program.attributeLocation("a_color");
    CNCVertex::program.enableAttributeArray(colorLocation);
    CNCVertex::program.setAttributeBuffer(colorLocation, GL_FLOAT, offset, 3, sizeof(CNCVertex));

    // Draw cube geometry using indices from VBO 1
    //glDrawElements(GL_TRIANGLES, indexLength, GL_UNSIGNED_INT, 0);
    glDrawElements(GL_TRIANGLES, g_mesh->render->indexCount, GL_UNSIGNED_SHORT, 0);

    CNCVertex::program.release();

    m_frame++;
}

double g_time = 0;


int main(int argc, char **argv){
    downloadLocation = QStandardPaths::locate(QStandardPaths::DownloadLocation, QString(), QStandardPaths::LocateDirectory);
    downloadLocation += "mesh.smls";

    QGuiApplication app(argc, argv);

    QSurfaceFormat format;
    format.setDepthBufferSize(24);
    QSurfaceFormat::setDefaultFormat(format);

    OpenGLWindow window;
    window.setFormat(format);
    window.resize(window_width, window_height);
    window.show();

    window.setAnimating(true);

    return app.exec();
}
//! [2]


//! [3]

//! [4]

